#!/usr/bin/env python
#
#
#            T3D2VTK : Convertor from T3D output into VTK legacy file         
#
#                            version 0.1 (2014-12-19)                         
#
#                       Copyright (C)  2014  Jan Stransky                     
#
#           Czech Technical University, Faculty of Civil Engineering,         
#       Department of Structural Mechanics, 166 29 Prague, Czech Republic     
#
#  T3D2VTK is free software: you can redistribute it and/or modify it under
#  the terms of the GNU Lesser General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  T3D2VTK is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
#  more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program. If not, see <http://www.gnu.org/licenses/>.
#

##############################################################################
# See T3D manual (http://mech.fsv.cvut.cz/~dr/software/T3d/guide/guide.html)
# and VTK legacy file manual (http://www.vtk.org/VTK/img/file-formats.pdf) for
# more information and better understanding od this script
##############################################################################
import sys
import os

# element degree
LINEAR           = 1
QUADRATIC        = 2
QUADRATIC_BUBBLE = 3


def printIntro(l=70):
	"""Prints intro"""
	print l*'_'
	print
	print 'T3D2VTK - Convertor from T3D output into VTK legacy file'.center(l)
	print '(C) 2014 Jan Stransky'.center(l)
	print l*'_'
	print

def printHelp():
	"""Prints help message"""
	print ' usage: [python] t3d2vtk.py t3dOutFileName vtkInFileName'
	print

class T3D2VTKError(Exception):
	"""User defined error"""
	def __init__(self, msg):
		self.msg = msg
	def __str__(self):
 		return repr(self.msg)

def readAndCheckArgs():
	# TODO test
	#if '--test' in sys.argv:
	#	do something
	if len(sys.argv) != 3:
		printHelp()
		sys.exit()
	iName,oName = sys.argv[1],sys.argv[2]
	if not os.path.exists(iName):
		raise IOError, "File '%s' does not exist"%iName
	try:
		f = open(oName,'w')
		f.close()
	except IOError:
		raise IOError, "File '%s' could not be opened for writing"%oName
	return iName,oName

def checkElemDegree(degree):
	if not degree in (LINEAR,QUADRATIC,QUADRATIC_BUBBLE):
		raise ValueError, "Element degree should be %s, %s or %s, not %s"%[str(v) for v in (LINEAR,QUADRATIC,QUADRATIC_BUBBLE,degree)]

class Node:
	"""Class representing FEM Node"""
	def __init__(self):
		self.id = -1
		self.property = 0
		self.x = self.y = self.z = 0.
	def fromLine(self,line):
		"""Reads and sets attributes from T3D file line"""
		ls = line.split()
		self.id,self.property = [int(ls[j]) for j in (0,6)]
		self.x,self.y,self.z = [float(ls[j]) for j in (1,2,3)]
		return self

class Element:
	"""Base class representing FEM element"""
	def __init__(self,degree):
		checkElemDegree(degree)
		self.id = -1
		self.property = 0
		self.nodes = None
		self.degree = degree
		self.numberOfNodes = -1
		self.vtkCellType = -1
	def fromLine(self,line):
		"""Reads and sets attributes from T3D file line"""
		ls = line.split()
		self.id = int(ls[0])
		self._setNodesAndProperty(ls)
		return self
	def _setNodesAndProperty(self,ls):
		self.nodes = [int(ls[i]) for i in xrange(1,self.numberOfNodes+1)]
		self.property = int(ls[self.numberOfNodes+3])
	def vtkNodes(self):
		"""Returns VTK nodes from T3D nodes"""
		raise NotImplementedError

class Edge(Element):
	"""Class representing edge element"""
	def __init__(self,degree):
		Element.__init__(self,degree)
		self.numberOfNodes = 2 if self.degree==LINEAR else 3
		self.vtkCellType = 3 if self.degree == LINEAR else 21
	def vtkNodes(self):
		return [n-1 for n in self.nodes]

class Triangle(Element):
	"""Class representing triangle element"""
	def __init__(self,degree):
		Element.__init__(self,degree)
		self.numberOfNodes = 3 if self.degree==LINEAR else 6 if self.degree==QUADRATIC else 7
		self.vtkCellType = 5 if self.degree == LINEAR else 22
	def vtkNodes(self):
		return [n-1 for n in self.nodes] if self.degree in (LINEAR,QUADRATIC) else [n-1 for n in self.nodes[:6]]

class Quadrilateral(Element):
	"""Class representing quadrilateral element"""
	def __init__(self,degree):
		Element.__init__(self,degree)
		self.numberOfNodes = 4 if self.degree==LINEAR else 8 if self.degree==QUADRATIC else 9
		self.vtkCellType = 9 if self.degree == LINEAR else 23
	def vtkNodes(self):
		return [n-1 for n in self.nodes] if self.degree in (LINEAR,QUADRATIC) else [n-1 for n in self.nodes[:8]]

class Tetrahedron(Element):
	"""Class representing tetrahedron element"""
	def __init__(self,degree):
		Element.__init__(self,degree)
		self.numberOfNodes = 4 if self.degree==LINEAR else 10 if self.degree==QUADRATIC else 15
		self.vtkCellType = 10 if self.degree == LINEAR else 24
	def vtkNodes(self):
		return [n-1 for n in self.nodes] if self.degree in (LINEAR,QUADRATIC) else [n-1 for n in self.nodes[:10]]

class Pyramid(Element):
	"""Class representing pyramid element"""
	def __init__(self,degree):
		Element.__init__(self,degree)
		self.numberOfNodes = 5 if self.degree==LINEAR else 13 if self.degree==QUADRATIC else 19
		self.vtkCellType = 14
	def vtkNodes(self):
		return [n-1 for n in self.nodes[:5]]

class Wedge(Element):
	"""Class representing wedge element"""
	def __init__(self,degree):
		Element.__init__(self,degree)
		self.numberOfNodes = 6 if self.degree==LINEAR else 15 if self.degree==QUADRATIC else 21
		self.vtkCellType = 13
	def vtkNodes(self):
		return [self.nodes[i]-1 for i in (0,2,1,3,5,4)]

class Hexahedron(Element):
	"""Class representing hexahedron element"""
	def __init__(self,degree):
		Element.__init__(self,degree)
		self.numberOfNodes = 8 if self.degree==LINEAR else 20 if self.degree==QUADRATIC else 27
		self.vtkCellType = 12 if self.degree == LINEAR else 25
	def vtkNodes(self):
		return [n-1 for n in self.nodes] if self.degree in (LINEAR,QUADRATIC) else [n-1 for n in self.nodes[:20]]



def readInputFile(fName):
	"""Reads input file (T3D output file :-) and returns nodes and elements"""
	f = open(fName)
	#
	nNodes,nEdges,nTrias,nQuads,nTetras,nPyrams,nWedges,nHexas = 0,0,0,0,0,0,0,0
	nodes,edges,trias,quads,tetras,pyrams,wedges,hexas = [],[],[],[],[],[],[],[]
	meshType,elemDegree,renumType,outputType = [int(w) for w in f.readline().split()]
	if meshType == 3:
		nNodes,nEdges,nTrias,nTetras = [int(w) for w in f.readline().split()]
	if meshType == 4:
		nNodes,nEdges,nQuads,nHexas = [int(w) for w in f.readline().split()]
	if meshType == 7:
		nNodes,nEdges,nTrias,nQuads,nTetras,nPyrams,nWedges,nHexas = [int(w) for w in f.readline().split()]
	else:
		raise T3D2VTKError, "Corrupted T3D output file (?), mesh type should be 3, 4 or 7, not %s"%str(meshType)
	#
	if nNodes:
		f.readline()
		nodes = [Node().fromLine(f.readline()) for i in xrange(nNodes)]
	if nEdges:
		f.readline()
		edges = [Edge(elemDegree).fromLine(f.readline()) for i in xrange(nEdges)]
	if nTrias:
		f.readline()
		trias = [Triangle(elemDegree).fromLine(f.readline()) for i in xrange(nTrias)]
	if nQuads:
		f.readline()
		quads = [Quadrilateral(elemDegree).fromLine(f.readline()) for i in xrange(nQuads)]
	if nTetras:
		f.readline()
		tetras = [Tetrahedron(elemDegree).fromLine(f.readline()) for i in xrange(nTetras)]
	if nPyrams:
		f.readline()
		pyrams = [Pyramid(elemDegree).fromLine(f.readline()) for i in xrange(nPyrams)]
	if nWedges:
		f.readline()
		wedges = [Wedge(elemDegree).fromLine(f.readline()) for i in xrange(nWedges)]
	if nHexas:
		f.readline()
		hexas = [Hexa(elemDegree).fromLine(f.readline()) for i in xrange(nHexas)]
	f.close()
	elems = edges + trias + quads + tetras + pyrams + wedges + hexas
	return nodes,elems

def writeOutputFile(fName,nodes,elems):
	"""Converts nodes and elements into VTK legacy file"""
	f = open(fName,'w')
	f.write('# vtk DataFile Version 2.0\ncomment\nASCII\nDATASET UNSTRUCTURED_GRID\n')
	#
	f.write('\nPOINTS %d double\n'%len(nodes))
	f.writelines('%.8e %.8e %.8e\n'%(n.x,n.y,n.z) for n in nodes)
	#
	f.write('\nCELLS %d %d\n'%(len(elems),sum(1+len(e.vtkNodes()) for e in elems)))
	for e in elems:
		ns = e.vtkNodes()
		f.write('%d %s\n'%(len(ns),' '.join(str(n) for n in ns)))
	f.write('\nCELL_TYPES %d\n'%len(elems))
	f.writelines('%d\n'%e.vtkCellType for e in elems)
	#
	f.write('\nPOINT_DATA %d\nSCALARS node_property int 1\nLOOKUP_TABLE default\n'%len(nodes))
	f.writelines('%d\n'%n.property for n in nodes)
	#
	f.write('\nCELL_DATA %d\nSCALARS elem_property int 1\nLOOKUP_TABLE default\n'%len(elems))
	f.writelines('%d\n'%e.property for e in elems)
	f.close()


def main():
	printIntro()
	iName,oName = readAndCheckArgs()
	nodes,elems = readInputFile(iName)
	writeOutputFile(oName,nodes,elems)

if __name__ == "__main__":
	main()
